/**
 * HMsweep - software to measure frequnecy responses with the sound card and a
 * Hameg digital storage oscilloscope
 *         
 * Copyright (C) 2007, 2008, 2016 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@opppf.de
 * http://opppf.de/HMsweep/
 */
#include <float.h>
#include <math.h>
#include <pthread.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <termios.h>
#include <unistd.h>
#include "osc_api.h"
#include "si_prefix.h"
#include "sinegen.h"
#include "sinegen_hm.h"

#define VERSION "1.11 2016-02-19"
#define SERIAL_PORT_OSC "/dev/ttyS0"
#define ALSA_DEVICE "default"
#define R_DEFAULT 1000
#define GNUPLOT "gnuplot"
#define LPR "|lpr"

/* constants */
#define HALFTONE 1.059463094

/* debugging */
#define debugc(x) fprintf( stderr, "%s=%c\n", #x, (x) )
#define debugi(x) fprintf( stderr, "%s=%d   %#06x\n", #x, (x), (x) )
#define debugd(x) fprintf( stderr, "%s=%g\n", #x, (x) )


/* ---------------------------------------------------------------------- 
 * typedefs
 * ---------------------------------------------------------------------- */


/**
 * Common arguments
 */
typedef struct
{
  int help;
  int version;
  int interactive;
  char output_mode;
  char * name;
} arg_t;


/**
 * Arguments for frequency sweep
 */
typedef struct
{
  double fmin;
  double fmax;
  double fstep;
  int fsteps;
  char mode;
  double R;
  char * serial;
  char * alsa;
  char * func;
  double genFmin;
  double genFmax;
} arg_sweep_t;


/**
 * output modes 
 */
#define OUT_PS 1
#define OUT_EPS 2
#define OUT_X11 4
#define OUT_PLT 8
#define OUT_DAT 16


/* ---------------------------------------------------------------------- 
 * little helpers
 * ---------------------------------------------------------------------- */

/** 
 * Determine number of halftones in intervall
 * i  intervall
 * return: number of halftones
 */
int halftones( double i )
{
  return (int)floor( 1 + fabs( log( i ) / log( HALFTONE ) ) );
}


/**
 * Determine step size for n frequencies in intervall [a,b].
 * a and b are explicitly _within_ the intervall.
 *
 * a	lowest frequency
 * b	highest frequency
 * n	number of steps.
 * return: facor for a single step
 */
double calc_fstep( double a, double b, int n )
{
  return pow( b / a, 1.0 / (n-1) );
}


/** Print error message and exit with code 1 */
void error( const char * e )
{
  fputs( e, stderr );
  fputs( "\nHelp with -h.\n", stderr );
  exit( 1 );
}


/**
 * generate filename from sweep args.
 *
 * arg	sweep args
 * name basename for file
 * end	fiel ending
 * return: filename as string. Must be freed by user.
 */
char * make_filename( const arg_sweep_t * arg, const char * name, 
                      const char * end )
{
  int l;
  char * s;
  char * mode;

  l = strlen( name ) + 23 + strlen( end );
  s = malloc( l * sizeof(char) );
  switch ( arg->mode )
  {
    case 'a': mode = "abs"; break;
    case 'b': mode = "bode"; break;
    case 'i': mode = "imp"; break;
    default: error( "Illegal mode in make_filename()" );
  }
  sprintf( s, "%s_%.1f-%.1f_%s.%s", name, arg->fmin, arg->fmax, mode, end );
  return s;
}


/**
 * open gnuplot
 * 
 * per	1 = persistive, 0 = not persistive
 */
FILE * open_gnuplot( int per )
{
  FILE * gp;

  gp = popen( GNUPLOT, "w" );
  if ( gp == 0 )
  {
    fprintf( stderr, "Can't open '%s'.\n", GNUPLOT );
    exit( 1 );
  }
  fputs( "set mouse\n", gp );
  if ( per )
    fputs( "set terminal x11 0 persist\n", gp );
  return gp;
}


/**
 * close gnuplot 
 */
void close_gnuplot( FILE * gp )
{
  fputs( "quit\n", gp );
  fflush( gp );
  pclose( gp );
}


/**
 * block SIGINT and install igonre handler
 * Should be done before any tread is started (sinegen).
 */
void block_sig_int()
{
  sigset_t mask;
  struct sigaction action;

  /* block signal */
  sigemptyset( &mask );
  sigaddset( &mask, SIGINT );
  pthread_sigmask( SIG_BLOCK, &mask, NULL );
  /* install ignore hanler */
  action.sa_handler = SIG_IGN;
  action.sa_mask = mask;
  action.sa_flags = 0;
  sigaction( SIGINT, &action, NULL );
}


/**
 * unblock SIGINT
 */
void unblock_sig_int()
{
  sigset_t mask;

  sigemptyset( &mask );
  sigaddset( &mask, SIGINT );
  pthread_sigmask( SIG_UNBLOCK, &mask, NULL );
}


/**
 * chek if SIGINT is pending
 * return: 0 = not pending, 1 = pending, -1 = error
 */
int check_sig_int()
{
  sigset_t mask;
  int p;

  sigpending( &mask );
  p = sigismember( &mask, SIGINT );
  if ( p )
    fputs( "Got SIGINT, aborting.\n", stderr );
  return p;
}


/* ---------------------------------------------------------------------- 
 * vector math
 * ---------------------------------------------------------------------- */


/**
 * square(x)
 */
#define square(x) (x)*(x)


/**
 * sign(x)
 */
#define sign(x) (((x)<0)?-1:((x)>0?1:0))


/**
 * substracts mean value from buffer
 *
 * b		buffer
 * N		number of values
 */
void ch_sub_mean( double * b, int N )
{
  int i;
  double m;

  m = 0;
  for ( i = 0; i < N; i++ )
    m += b[i];
  m /= N;
  for ( i = 0; i < N; i++ )
    b[i] -= m;
}


/**
 * Calculate the inner product.
 *
 * a,b	two vectors, mean value zero
 * N	dimension of vectors
 * return: the inner product, |a| |b| cos(phi)
 *         |a| and |b| are the amplitudes of sine waves.
 */
double ch_inner_product( const double * a, const double * b, int N )
{
  int i;
  double ip = 0;

  for ( i = 0; i < N; i++ )
    ip += a[i] * b[i];
  ip *= 2.0 / N;
  return ip;
}


/**
 * claculate amplitude of (assumed) sine wave.
 *
 * b		buffer with mean free sine wave
 * N		Numer of samples in buffer
 * return: Amplitude
 */
double ch_amp( const double * b, int N )
{
  int i;
  double A;

  A = 0;
  for ( i = 0; i < N; i++ )
    A += square( b[i] );
  A = sqrt( 2 * A / N );
  return A;
}


/**
 * relative phase of two waves.
 *
 * a, b		two buffers containing a whole number of waves
 * A, B		Amplitude of sine waves.
 * Ns		number of samples per buffer
 * Nw		number of whole waves in each buffers
 * return: phi in degree.
 *         If 'a' is cos an 'b' is sin then phi is +90.
 */
double ch_phase( const double * a, const double * b, double A, double B, 
                 int Ns, int Nw )
{
  double Re, Im;
  double N_one_wave;
  const double * b90;
  int N90;
  double phi;

  /* Imaginary part: shift 'a' from cos to sin
   *             <=> let 'a' begin 90 earlier
   *             <=> let 'b' begin 90 later
   * To avoid leaving the array, take one wave less.
   */
  N_one_wave = (int)floor( 0.5 + (double)Ns / (double)Nw );
  b90 = &b[ (int)(N_one_wave / 4) ];
  N90 = (int)floor( 0.5 + N_one_wave * ( Nw - 1 ) );
  Re = ch_inner_product( a, b, Ns );
  Im = ch_inner_product( a, b90, N90 );
  
  phi = acos( Re / A / B );
  phi *= sign( Im );            /* sign( asin(x) ) = sign( x ) */
  return 180 / M_PI * phi;
}


/* ---------------------------------------------------------------------- 
 * universal sine generator
 * ---------------------------------------------------------------------- */

typedef struct 
{
    int type; /* 'a' = alsa, 'f' = function generator */
    sinegen_alsa_t * alsa;
    sinegen_hm_t * hm;
} sinegen_t;


/**
 * start sine generator, determine type (alsa or func generator)
 *
 * sw_arg	sweep args, to determine type and device of generator
 * f		frequency to generate
 * return: handle for sine generator
 */
sinegen_t * sinegen_start( arg_sweep_t * sw_arg, double f )
{
    sinegen_t * ret = malloc( sizeof(sinegen_t) );

    if ( sw_arg->alsa != NULL )
    {
	ret->type = 'a';
	ret->alsa = sinegen_alsa_start( sw_arg->alsa, f );
    }
    else if ( sw_arg->func != NULL )
    {
	ret->type = 'f';
	ret->hm = sinegen_hm_start( sw_arg->func, f );
    }
    else
    {
	fprintf( stderr, "Error in sinegen_start. This is a bug!\n" );
	exit( 1 );
    }
    return ret;
}

/**
 * stop sine generator
 *
 * h	handle of sine generator, created by sinegen_start()
 */
void sinegen_stop( sinegen_t * h )
{
    if ( h->type == 'a' )
	sinegen_alsa_stop( h->alsa );
    if ( h->type == 'f' )
	sinegen_hm_stop( h->hm );
}


/* ---------------------------------------------------------------------- 
 * sweep
 * ---------------------------------------------------------------------- */


/** 
 * Calculate best timescale for given frequncy.
 * There must be at least 2 waves on the screen.
 *
 * f	frequency
 * ts	optimal timescale (int for osc_set_tb() )
 * Nw	number of whole waves in record.
 * Ns	number of samples in Nw waves.
 */
void calc_best_ts( double f, int * ts, int * Nw, int * Ns )
{
  double topt;
  int ti;

  topt = 0.2 / f;
  for ( ti = 0; OSC_s_DIV[ti] < topt; ti++ )
    if ( ti == OSC_s_DIVN - 1 )
      break;
  *ts = ti;
  /* whole waves on screen */
  *Nw = floor( 2048.0 / 2000.0 * 10.0 * OSC_s_DIV[ti] * f ); 
  *Ns = (int)floor( 0.5 + *Nw / f * 200.0 / OSC_s_DIV[ti] );
  if ( *Nw < 2 )
    error( "Waves do not fit in scope screen." );
  return;
}


/* flags for check_amplitude */
#define Y_OK 1000
#define Y_FAULT 2000


/**
 * Check size of signal amplitude.
 *
 * ch	channel data
 * N	number of samples
 * Yi	present y-scale as index
 * return: best y-scle as index
 *         Y_OK if present index is the best
 *         best y-scale + Y_OK if signal is really too small/large
 */
int check_amplitude( double * ch, int N, int Yi )
{
  int i;
  double peak = 0;
  double mean = 0;
  int clip = 0;
  const double ymax = 5.12 * OSC_V_DIV[Yi]; /* 25 LSB/DIV, 8 bit */

  for ( i = 0; i < N; i++ )
    mean += ch[i];
  mean /= N;
  for ( i = 0; i < N; i++ )
  {
    if ( fabs( ch[i] - mean ) > peak )
      peak = fabs( ch[i] - mean );
    if ( fabs( ch[i] ) >= ymax )
      clip++;
  }
  if ( clip && Yi == OSC_V_DIVN )
    fputs( "WARING! Signal too large!\n", stderr );
  /* If signal is clipped the peak value has to be guessed.
   * 
   * linear guess:
   *
   *         ggCCC      CCgggCCCCCgg       (g)ood / (C)lipped
   * peak _ _ _  .     .
   *            /:      \
   * ymax - - -/-:      -\- - - - - - -
   *          /: :        \
   *      ---/----      ---\-------/----
   *         : : :          \     /
   *         : :x:        - -\- -/- - -
   *         :   :            \ /
   *         : n :             '
   * 
   * N    = number of samples
   * n    = N / 4
   * clip = number of clipped samples
   * x    = clip / 4
   *
   * regarding one quaterwave:
   *
   *  peak        N
   * ------ = ----------
   *  ymax     N - clip
   */
  if ( clip )
    peak = ymax * N / ( N - clip );
  /* determine best yscale */
  for ( i = 0; OSC_V_DIV[i] * 2.5 < peak; i++ )
    if ( i == OSC_V_DIVN - 1 )
      break;
  /* return scale suggestion */
  if ( i == Yi )
    return Y_OK;
  if ( clip )
    i += Y_FAULT;
  else 
    if ( peak < 0.5 * OSC_V_DIV[Yi] ) 
      i += Y_FAULT;
  return i;
}


/** 
 * start measurement, wait, and hold signal
 */
void measure_and_wait( OSC * o )
{
  struct timeval wait;
  double acq_time;
  int acq_sec, acq_usec;

  acq_time = 10 * OSC_s_DIV[osc_get_tb_scale( o, 'A', NULL )];
  acq_sec = (int)floor( acq_time );
  acq_usec = (int)( 1e6 * ( acq_time - floor( acq_time ) ) );
  /* 0.5 s + acquisition time */
  wait.tv_sec = acq_sec;
  wait.tv_usec = 500000 + acq_usec;
  osc_set_hold( o, 0 );
  select( 0, NULL, NULL, NULL, &wait );
  osc_set_hold( o, 1 );
  return;
}


/**
 * Do the sweep.
 * Control the sine generator, get data from osc, calculate level and phase.
 *
 * sw_freq, sw_abs and sw_phase become allocated and must be freed by user.
 * sw_freq, sw_abs and sw_phase are of length sw_arg->steps.
 *
 * The signal SIGINT should be blocked. I one is pending, the sweep aborts.
 *
 * sw_arg	arguments for sweep
 * sw_freq	frequency
 * sw_abs	absolute value
 * sw_phase	phase of value (may be NULL for absolute)
 * return: 0 = finished, else aborted
 */
int sweep( arg_sweep_t * sw_arg, double ** sw_freq, double ** sw_abs, 
           double ** sw_phase )
{
  OSC * o;
  sinegen_t * sinegen;
  double * ch1;
  double * ch2;
  double A1, A2; /* Amplitude for each channel */
  int sY1, sY2;  /* suggest changed y-scale, see check_amplitude() */ 
  double f;      /* present frequency */
  int ts;        /* optimal timescale */
  int Ns;        /* samples from scope containing an whole number of waves */
  int Nw;        /* Number of whole waves on screen */
  int abort = 0; /* flag indicating abortion by SIGINT */
  int i;

  /* init */
  ch1 = malloc( 2048 * sizeof(double) );
  ch2 = malloc( 2048 * sizeof(double) );
  *sw_freq = malloc( sw_arg->fsteps * sizeof(double) );
  *sw_abs = malloc( sw_arg->fsteps * sizeof(double) );
  *sw_phase = malloc( sw_arg->fsteps * sizeof(double) );
  block_sig_int();
  o=osc_open( sw_arg->serial );
  if ( o->error[0] ) error( "Can not open serial port for scope." );
  /* set osc to appropriate state */
  sinegen = sinegen_start( sw_arg, sw_arg->fmin ); /* for autoset */
  osc_set_hormode( o, OSC_HOR_STORE | OSC_HOR_TB_a, 
    OSC_HOR_CT | OSC_HOR_XY | OSC_HOR_x10 | OSC_HOR_STORE | OSC_HOR_TB );
  osc_set_vermode( o, (sw_arg->mode=='a') ? OSC_VER_TRS_CH1 : OSC_VER_TRS_CH2, 
    OSC_VER_ALTTRG | OSC_VER_CHOP | OSC_VER_ADD | OSC_VER_TRS );
  osc_set_trigger( o, 0, OSC_TRI_COUPL_LF ); /* TODO: check scope type */
  osc_set_ch( o, 1, OSC_CHx_ON | OSC_CHx_AC, -1, 0xff );
  osc_set_ch( o, 2, (sw_arg->mode=='a')?0:OSC_CHx_ON | OSC_CHx_AC, -1, 0xff );
  osc_set_hold( o, 0 );
  osc_set_auto( o );
  sinegen_stop( sinegen );
    /* correct trigger source after autoscale*/
  osc_set_vermode( o, (sw_arg->mode=='a') ? OSC_VER_TRS_CH1 : OSC_VER_TRS_CH2, 
    OSC_VER_ALTTRG | OSC_VER_CHOP | OSC_VER_ADD | OSC_VER_TRS );
    /* correct Y positions after autoscale */
  osc_set_ypos( o, 1, 0 );
  osc_set_ypos( o, 2, 0 );
    /* correct Y amplitudes after autoscale */
  if ( sw_arg->mode != 'a' )
  {
    sY1 = osc_get_ch_scale( o, 1, NULL );
    sY2 = osc_get_ch_scale( o, 2, NULL );
    if ( sY1 > 0 )
      osc_set_ch( o, 1, -1, --sY1, 0xff );
    if ( sY2 > 0 )
      osc_set_ch( o, 2, -1, --sY2, 0xff );
  }
  /* sweep loop */
  sY1 = sY2 = Y_OK;
  f = sw_arg->fmin;
  for ( i = 0; i < sw_arg->fsteps; i++ )
  {
    if ( (abort = check_sig_int()) )
      break;
    fprintf( stderr, "f%-3i = %7.1f Hz\n", i, f );
    calc_best_ts( f, &ts, &Nw, &Ns );
    osc_set_tb( o, 'A', -1, ts, 0 );
    if ( sY1 != Y_OK )
      osc_set_ch( o, 1, -1, sY1, 0xff );
    if ( sY2 != Y_OK )
      osc_set_ch( o, 2, -1, sY2, 0xff );
    /* measure until data is valid */
    while ( 1 ) 
    {
      sinegen = sinegen_start( sw_arg, f );
      measure_and_wait( o );
      sinegen_stop( sinegen );
      osc_read_chan( o, 1, 0, Ns, ch1 );
      sY1 = check_amplitude( ch1, Ns, osc_get_ch_scale( o, 1, NULL ) );
      if ( sY1 >= Y_FAULT ) /* signal too small/large */
      {
	sY1 -= Y_FAULT;
	fprintf( stderr, "Repeatng CH1 with %.3f V/DIV.\n", OSC_V_DIV[sY1] );
        osc_set_ch( o, 1, -1, sY1, 0xff );
	sY1 = Y_OK;
	continue;
      }
      if ( sw_arg->mode != 'a' )
      {
	osc_read_chan( o, 2, 0, Ns, ch2 );
	sY2 = check_amplitude( ch2, Ns, osc_get_ch_scale( o, 2, NULL ) );
	if ( sY2 >= Y_FAULT ) /* signal too small/large */
	{
	  sY2 -= Y_FAULT;
	  fprintf( stderr, "Repeatng CH2 with %.3f V/DIV.\n", OSC_V_DIV[sY2] );
	  osc_set_ch( o, 2, -1, sY2, 0xff );
	  sY2 = Y_OK;
	  continue;
	}
      }
      break;
    }
    /* write data to freq, abs, and phase array */
    (*sw_freq)[i] = f;
    switch ( sw_arg->mode )
    {
      case 'a':
	ch_sub_mean( ch1, Ns );
	(*sw_abs)[i] = 20 * log10( ch_amp( ch1, Ns ) );
	break;
      case 'b':
	/* Ue = ch1, Ua = ch2 */
	ch_sub_mean( ch1, Ns ); ch_sub_mean( ch2, Ns );
	A1 = ch_amp( ch1, Ns ); A2 = ch_amp( ch2, Ns );
	(*sw_abs)[i] = 20 * log10( A1 / A2 );
	(*sw_phase)[i] = ch_phase( ch2, ch1, A2, A1, Ns, Nw );
	break;
      case 'i':
	/* U = ch1, I = -ch2 / R    ***    -1 -> PHI + 180 */
	ch_sub_mean( ch1, Ns ); ch_sub_mean( ch2, Ns );
	A1 = ch_amp( ch1, Ns ); A2 = ch_amp( ch2, Ns );
	(*sw_abs)[i] =  A1 / A2 * sw_arg->R;
	(*sw_phase)[i] = 180 + ch_phase( ch1, ch2, A1, A2, Ns, Nw );
	if ( (*sw_phase)[i] > 180 ) /* [0..360] -> [-180..180] */
	  (*sw_phase)[i] -= 360;
	break;
    }
    f *= sw_arg->fstep;
  }
  /* clean up */
  osc_close( o );
  free( ch1 );
  free( ch2 );
  unblock_sig_int();
  return abort;
}


/* ---------------------------------------------------------------------- 
 * Data output --  Gnuplot
 * ---------------------------------------------------------------------- */


/**
 * Determine maximum value in array
 * a	array
 * N	length
 * return: maximum vlaue
 */
double maxval( const double * a, int N )
{
  int i;
  double max = -DBL_MAX;

  for ( i = 0; i < N; i++ )
    if ( a[i] > max )
      max = a[i];
  return max;
}


/**
 * Determine minimum value in array
 * a	array
 * N	length
 * return: minimum vlaue
 */
double minval( const double * a, int N )
{
  int i;
  double min = DBL_MAX; /* is MAXDOUBLE in values.h */

  for ( i = 0; i < N; i++ )
    if ( a[i] < min )
      min = a[i];
  return min;
}


/**
 * Generate nice scale values.
 * Finds a step near s with _one_ significant digit in {1, 2, 5}.
 * Finds nearby values for beginning and end of scale.
 *
 * b		begin value
 * e		end value
 * n		number of steps
 * nb		nice begin
 * ne 		nice end
 * ni		nice intervall
 */
void make_nice_scale( double b, double e, double n, 
                      double * nb, double * ne, double * ni )
{
  double step = ( e - b ) / n;
  double exponent = floor( log10( step ) );
  double mantisse = step / pow( 10, exponent );

  /* nice increment */
  *ni = 1;
  if ( mantisse > 1.3 && mantisse <= 3 )
    *ni = 2;
  if ( mantisse > 3 && mantisse <= 7 )
    *ni = 5;
  if ( mantisse > 7 )
    *ni = 10;
  *ni *= pow( 10, exponent );
  /* nice boundaries */
  *nb = *ni * floor( b / *ni );
  *ne = *ni * ceil( e / *ni );
}


/**
 * Nice labels for lagscale if only one or two decades are shown.
 * This function is needed due to the weak logscale features of gnuplot.
 *
 * min		minimum value
 * max		maximum value
 * gp		file to write the gnuplot commands to
 */
void manual_logscale( FILE * gp, double min, double max )
{
  double first, last, step;
  double i;
  int ma, ex, scale;
  char s[4];

  fputs( "set xtics (", gp );
  /* apply linear distributed tics */
  if ( max / min < 10 )
  {
    make_nice_scale( min, max, 4, &first, &last, &step );
    for ( i = first; i <= last; i += step )
    {
      fprintf( gp, " %g", i ); 
      if ( i != last )
	fprintf( gp, ", '' %g,", i + 0.5*step );
    }
  }
  /* apply 1-3 or 1-2-5 scale */
  else
  {
    /* 1-2-5 scale */
    if ( max/min < 60 )
      scale = 125;
    else
      scale = 13;
    for ( ex = floor(log10(min)); ex < ceil(log10(max)); ex++ )
      for ( ma = 1; ma < 10; ma++ )
      {
	strcpy( s, "'' " );
	switch ( ma )
	{
	  case 1: s[0] = 0; break;
	  case 2: if ( scale == 125 ) s[0] = 0; break;
	  case 3: if ( scale ==  13 ) s[0] = 0; break;
	  case 5: if ( scale == 125 ) s[0] = 0; break;
	}
	fprintf( gp, " %s%de%d,", s, ma, ex );
      }
    fprintf( gp, " 1e%d", ex );
  }
  fputs( " )\n", gp );
}


/**
 * different format for X11 and ps/eps
 */
#define onX(x, y) ((format==OUT_X11)?(x):(y))


/**
 * write commands to gnuplot
 * 
 * gp		file handle for gnuplot
 * arg		args for sweep
 * abslt	absolute values
 * phase	phase values, NULL for mode abs.
 * fname	filename for gnuplot output, NULL for X11
 * format	output format, OUT_PS, OUT_EPS or OUT_X11
 */
void write_to_gnuplot( FILE * gp, arg_sweep_t * arg, const double * freq, 
     const double * abslt, const double * phase, char * fname, int format )
{
  double max_abslt, min_abslt, d_abslt, max_phase, min_phase, d_phase;
  double xmax, xmin, ymax, ymin, y2max, y2min;
  double first_tic, last_tic, tic_step;
  int i;

  /* header */
  fputs( "#!/usr/bin/gnuplot -persist\n", gp );
  fputs( "# GNUplot script made by HMsweep\n", gp );
  fputs( "reset\n", gp );
  fputs( "\n", gp );
  /* format specific header */
  switch ( format )
  {
    case OUT_PS:
      fputs( "set terminal postscript landscape enhanced dl 3 lw 3 24\n", gp );
      fprintf( gp, "set output '%s'\n", fname );
      fputs( "set grid xtics mxtics ytics mytics y2tics my2tics lt 1 lw 0.7, lt 1 lw 0.3\n", gp );
      break;
    case OUT_EPS:
      fputs( "set terminal postscript eps enhanced dl 3 lw 3 31\n", gp );
      fprintf( gp, "set output '%s'\n", fname );
      fputs( "set grid xtics mxtics ytics mytics y2tics my2tics lt 1 lw 0.7, lt 1 lw 0.3\n", gp );
      break;
    case OUT_X11:
      fputs( "set terminal x11 0 title 'Monitor HM'\n", gp );
      fputs( "set grid xtics mxtics ytics mytics y2tics my2tics lt 7 lw 2, lt 7 lw 1\n", gp );
      break;
    default:
      error( "Illegal format in write_to_gnuplot." );
  }
  /* common header */
  fputs( "set encoding iso_8859_1\n", gp );
  fputs( "set decimalsign ','\n", gp );
  fputs( "set style data lines\n", gp );
    /* x axis */
  xmin = 0.9 * ( arg->fmin < arg->fmax)?arg->fmin:arg->fmax;
  xmax = 1.1 * ( arg->fmin < arg->fmax)?arg->fmax:arg->fmin;
  if ( xmax / xmin < 100 )
    manual_logscale( gp, xmin, xmax );
  fputs( "set xlabel 'f in Hz'\n", gp );
  fprintf( gp, "set xrange [%.0f:%.0f]\n", xmin, xmax );
  fputs( "set logscale x\n", gp );
  /* mode specific header */
    /* y axis */
  switch ( arg->mode )
  {
    case 'a':
      fputs( "set mytics 2\n", gp );
      if ( format == OUT_X11 )
	fputs( "set ylabel 'A in dBV'\n", gp );
      else
      {
	fputs( "set label 'A in dBV' at screen 0, 0.97\n", gp );
	fputs( "set tmargin 2\n", gp );
	fputs( "set bmargin 2\n", gp );
	fputs( "set rmargin 4.5\n", gp );
	fputs( "set lmargin 4\n", gp );
      }
      break;
    case 'b':
      max_abslt = maxval( abslt, arg->fsteps );
      min_abslt = minval( abslt, arg->fsteps );
      make_nice_scale( min_abslt, max_abslt, 8, 
		       &first_tic, &last_tic, &tic_step );
      d_abslt = last_tic - first_tic;
      ymax = last_tic;
      ymin = first_tic - 0.5 * d_abslt;
      fprintf( gp, "set ytics nomirror %g, %g, %g\n", 
		   first_tic, tic_step, last_tic  );
      fprintf( gp, "set yrange [%g:%g]\n", ymin, ymax );
      fputs( "set mytics 2\n", gp );
      if ( format == OUT_X11 )
	fputs( "set ylabel 'A in dB'\n", gp );
      else
      {
	fputs( "set label 'A in dB' at screen 0, 0.97\n", gp );
	fputs( "set tmargin 2\n", gp );
	fputs( "set bmargin 2\n", gp );
	fputs( "set rmargin 4.5\n", gp );
	fputs( "set lmargin 4\n", gp );
      }
      break;
    case 'i':
      max_abslt = maxval( abslt, arg->fsteps );
      min_abslt = minval( abslt, arg->fsteps );
      last_tic = pow( 10, ceil( log10( max_abslt ) ) );
      first_tic = pow( 10, floor( log10( min_abslt ) ) );
      d_abslt = last_tic / first_tic;
      ymax = last_tic;
      ymin = first_tic / sqrt( d_abslt );
      fputs( "set logscale y\n", gp );
      fprintf( gp, "set ytics nomirror %g, 10, %g\n", first_tic, last_tic  );
      fprintf( gp, "set yrange [%g:%g]\n", ymin, ymax );
      fputs( "set format y '10^{%L}'\n", gp );
      if ( format == OUT_X11 )
	fputs( "set ylabel 'Z in Ohm'\n", gp );
      else
      {
	fputs( "set label 'Z in {/Symbol W}' at screen 0.0, 0.97\n", gp );
	fputs( "set tmargin 2\n", gp );
	fputs( "set bmargin 2\n", gp );
	fputs( "set rmargin 4.5\n", gp );
	fputs( "set lmargin 4\n", gp );
      }
      break;
    default:
      error( "Illegal mode in write_to_gnuplot." );
  }
    /* y2 axis */
  if ( arg->mode != 'a' )
  {
    max_phase = 90 * ceil( maxval( phase, arg->fsteps ) / 90 );
    min_phase = 90 * floor( minval( phase, arg->fsteps ) / 90 );
    d_phase = max_phase - min_phase;
    y2max = min_phase + 3.3 * d_phase;
    y2min = min_phase;
    fprintf( gp, "set y2range [%g:%g]\n", y2min, y2max );
    fprintf( gp, "set y2tics nomirror %g, %g, %g\n", min_phase, 
             (d_phase<180)?30.0:((d_phase>180)?90.0:45.0), max_phase  );
    fputs( "set my2tics 3\n", gp );
    if ( format == OUT_X11 )
      fputs( "set y2label 'phi in '\n", gp );
    else
      fputs( "set label '{/Symbol f} in ' at screen 1.00, 0.48 right\n", gp );
    fprintf( gp, "set arrow from %g, second %g to %g, second %g nohead lt -1\n",
             xmin, max_phase + 0.05 * d_phase, 
	     xmax, max_phase + 0.05 * d_phase );
  }
  /* plot */
  fprintf( gp, "plot '-' notitle lw %d", onX( 2, 1 ) );
  if ( arg->mode != 'a' )
    fprintf( gp,  ", '-' axis x1y2 notitle lt %d lw %d\n", 
             onX( 3, 1 ), onX( 2, 1 ) );
  else
    fputs( "\n", gp );
  /* data */
  for ( i = 0; i < arg->fsteps; i++ )
    fprintf( gp, "%e  %e\n", freq[i], abslt[i] );
  fputs( "e\n", gp );
  if ( arg->mode != 'a' )
  {
    for ( i = 0; i < arg->fsteps; i++ )
      fprintf( gp, "%e  %e\n", freq[i], phase[i] );
    fputs( "e\n", gp );
  }
  /* close file */
  if ( format & ( OUT_PS | OUT_EPS ) )
    fputs( "set output\n", gp );
}


/**
 * write data as DAT file.
 * 
 * arg		args for sweep
 * abslt	absolute values
 * phase	phase values, may be NULL for mode abs.
 * bname	basename for file
 */
void write_dat( arg_sweep_t * arg, const double * freq, const double * abslt, 
               const double * phase, char * bname )
{
  int i;
  char * fname;
  FILE * f;

  fname = make_filename( arg, bname, "dat" );
  f = fopen( fname, "w" );
  switch ( arg->mode  )
  {
    case 'b':
      fprintf( f, "# %8s  %10s  %10s\n", "f [Hz]", "|A| [dB]", "phi []" );
      for ( i = 0; i < arg->fsteps; i++ )
        fprintf( f, "%10.3f  %10.2f  %10.1f\n", freq[i], abslt[i], phase[i] );
      break;
    case 'a':
      fprintf( f, "# %8s  %10s\n", "f [Hz]", "|A| [dBV]" );
      for ( i = 0; i < arg->fsteps; i++ )
	fprintf( f, "%10.3f  %10.2f\n", freq[i], abslt[i] );
      break;
    case 'i':
      fprintf( f, "# %8s  %10s  %10s\n", "f [Hz]", "Z [Ohm]", "phi []" );
      for ( i = 0; i < arg->fsteps; i++ )
        fprintf( f, "%10.3f  %10g  %10.1f\n", freq[i], abslt[i], phase[i] );
      break;
    default:
      error( "Unknown mode in write_data." );
  }
  fclose( f );
  free( fname );
}



/* ---------------------------------------------------------------------- 
 * interactive mode
 * ---------------------------------------------------------------------- */


/* buffer length for strings used in interactive mode */
#define INTBUFLEN 256


/**
 * reads one char unbuffered from stdin
 *
 * returns -1 on error
 */
char readchar()
{
  char ch;
  int n;
  static struct termios Otty, Ntty;

  fflush( stdout );
  tcgetattr( 0, &Otty );
  Ntty = Otty;
  /* unbuffered settings */
  Ntty.c_iflag = 0;
  Ntty.c_oflag = 0;
  Ntty.c_lflag &= ~ICANON; /* switch off line buffer */
  Ntty.c_lflag |= ECHO;  /* enable echo */
  Ntty.c_cc[VMIN]  = 1; /* minimum chars to wait for */
  Ntty.c_cc[VTIME] = 255; /* minimum wait time (ds). Does value matter? */
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Ntty) )
    return -1;
  n = read( 0, &ch, 1 ); 
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Otty) )
    return -1;
  if ( n != 1 )
    return -1;
  return (int)ch;
}


/**
 * read a line from f to buffer b with length l
 * If string in b ends on NL, replace it by \0.
 */
char * readline( FILE * f, char * b, size_t l )
{
  char * r;
  size_t n;

  r = fgets( b, l, f );
  if ( r ) 
  {
    n = strlen( b ) - 1;
    if ( b[n] == '\n') 
      b[n] = '\0';
  }
  return r;
}


/**
 * Ask for string
 *
 * give prompt "q [buf]> "
 * On empty input buffer is unchanged.
 * assumed buffer length is INTBUFLEN.
 *
 * q    question
 * buf  buffer to read to
 */
void int_ask_str( const char * q, char * buf )
{
  char line[INTBUFLEN] = "";

  fprintf( stdout, "%s [%s]> ", q, buf );
  fflush( stdout );
  readline( stdin, line, INTBUFLEN );
  if ( line[0] )
    strcpy( buf, line );
}


/**
 * Print selection box with braces.
 * 
 * The string s is printed, 
 * either in [] for ( a == b )
 * or in spaces for ( a != b )
 */
void int_print_selsction( FILE * f, const char * s, int a, int b )
{
  if ( a == b ) 
    fprintf( f, "\e[1m[%s]\e[m", s ); /* ANSII ESC */
  else
    fprintf( f, " %s ", s );
}


/**
 * print prompt for interactive mode
 */
void prompt( FILE * f, arg_sweep_t * arg, const char * message )
{
  fputs( "\e[H\e[J", stdout ); /* ANSII ESC */
  fprintf( f, "message :  %s\n\n", message );
  /* frequencies */

  fprintf( f, "sweep   :  f_miN =%12s", 
	   format_si( arg->fmin, si_prefix(arg->fmin,6), "Hz" ) );
  fprintf( f,"   f_maX =%12s\n", 
           format_si( arg->fmax, si_prefix(arg->fmax,6), "Hz" ) );
  /* steps */
  fprintf( f, "           Steps:  %4i", arg->fsteps );
  fputs( "                    Halftones\n", f );
  /* mode */
  fputs(  "mode    : ", f );
  int_print_selsction( f, "Abs", 'a', arg->mode );
  int_print_selsction( f, "Bode", 'b', arg->mode );
  int_print_selsction( f, "Impedance", 'i', arg->mode );
  fprintf( f, "     R =%12s\n", 
           format_si( arg->R, si_prefix(arg->R,3), "Ohm" ) );
  /* save as */
  fputs( "save as :  Ps        Eps        Gnuplot         Data\n", f );
  fputs( "           Quit           prinT            <-' sweep\n", f );
  /* prompt */
  fputs( "\nHMsweep > ", f );
  fflush( f );
}


/**
 * Initialise gnuplot, generate greeting screen
 */
void gnuplot_welcome( FILE * gp )
{
  fputs( "set mouse\n", gp );
  fputs( "set terminal x11 0 title 'Monitor HM'\n", gp );
  fputs( "set format ''\n", gp );
  fputs( "set zeroaxis\n", gp );
  fputs( "set label 'HMsweep' at 0, 0.3 center\n", gp );
  fputs( "\n", gp );
  fputs( "plot [-3.1415:3.1415][-1.1:1.1] sin(x) notitle\n", gp );
  fflush( gp );
}


/* Standard message for no error */
#define READY "READY."

/**
 * main loop for interactive mode
 */
void interactive( arg_sweep_t * arg )
{
  char message[INTBUFLEN] = READY;
  char base_name[INTBUFLEN] = "";
  char printer[INTBUFLEN] = LPR;
  char answer[INTBUFLEN];
  double * freq = NULL;
  double * abslt = NULL;
  double * phase = NULL;
  double * old_f;
  double * old_a;
  double * old_p;
  char * fname;
  char * fname2;
  FILE * gp;
  FILE * o;
  int c;

  gp = open_gnuplot( 0 );
  gnuplot_welcome( gp );
  do
  {
    prompt( stdout, arg, message );
    c = readchar();
    fputs( "\n", stdout );
    switch ( c )
    {
      case 'n': /* miN frequency */
	strcpy( message, READY );
	sprintf( answer, "%.2f", arg->fmin );
        int_ask_str( "f_min", answer );
        arg->fmin = scan_si( answer );
	if ( arg->fmin < arg->genFmin )
	{
	  arg->fmin = arg->genFmin;
	  snprintf( message, INTBUFLEN, "f_min must be at least %g.", 
		    arg->genFmin );
	}
	if ( arg->fmin > arg->genFmax )
	{
	  arg->fmin = arg->genFmax;
	  snprintf( message, INTBUFLEN, "f_min must be not larger than %g.", 
		    arg->genFmax );
	}
	arg->fstep = calc_fstep( arg->fmin, arg->fmax, arg->fsteps ); 
        break; 
      case 'x': /* maX frequency */
	strcpy( message, READY );
	sprintf( answer, "%.2f", arg->fmax );
        int_ask_str( "f_max", answer );
        arg->fmax = scan_si( answer );
	if ( arg->fmax < arg->genFmin )
	{
	  arg->fmax = arg->genFmin;
	  snprintf( message, INTBUFLEN, "f_max must be at least %g.", 
		    arg->genFmin );
	}
	if ( arg->fmax > arg->genFmax )
	{
	  arg->fmax = arg->genFmax;
	  snprintf( message, INTBUFLEN, "f_max must be not larger than %g.", 
		    arg->genFmax );
	}
	arg->fstep = calc_fstep( arg->fmin, arg->fmax, arg->fsteps ); 
        break; 
      case 's': /* Steps in sweep */
	strcpy( message, READY );
	sprintf( answer, "%i", arg->fsteps );
        int_ask_str( "Number of steps", answer );
        arg->fsteps = atoi( answer );
	arg->fstep = calc_fstep( arg->fmin, arg->fmax, arg->fsteps ); 
	if ( fabs( arg->fstep - 1 ) < 0.002 )
        snprintf( message, INTBUFLEN,
	         "%i steps from %.1f Hz to %.1f Hz are pretty much.", 
		  arg->fsteps, arg->fmin, arg->fmax );
	if ( arg->fsteps < 1 )
	{
	  arg->fsteps = 1;
	  strcpy( message, "There must be at least one step." );
	}
        break; 
      case 'h': /* set Halftones */
	arg->fsteps = halftones( arg->fmax / arg->fmin );
	arg->fstep = calc_fstep( arg->fmin, arg->fmax, arg->fsteps ); 
	strcpy( message, READY );
        break;
      case 'a': /* set mode: Absolute Bode Impedance */
      case 'b':
      case 'i':
	arg->mode = c;
	strcpy( message, READY );
        break;
      case 'r': /* Resistor */
	strcpy( message, READY );
	sprintf( answer, "%7.4g", arg->R );
        int_ask_str( "Resistor for current measurement", answer );
        arg->R = scan_si( answer );
	if ( arg->R == 0 )
	{
	  arg->R = 1;
	  strcpy( message, "Resistor must not be zero." );
	}
	break;
      case '\r': /* do the sweep */
      case '\n':
	strcpy( message, READY );
	if ( arg->fmin == arg->fmax )
	{
	  strcpy( message, "f_min and f_max must differ." );
	  break;
	}
	old_f = freq; old_a = abslt; old_p = phase;
	puts( "Sweeping. Abort with ^C." );
	if ( sweep( arg, &freq, &abslt, &phase ) != 0 ) /* sweep aborted */
	{
	  free( freq ); free( abslt ); free( phase );
	  freq = old_f; abslt = old_a; phase = old_p;
	  strcpy( message, "Sweep aborted." );
	  break;
	}
	free( old_f ); free( old_a ); free( old_p );
	write_to_gnuplot( gp, arg, freq, abslt, phase, NULL, OUT_X11 );
	fflush( gp );
        break;
      case 'p': /* save as postscript */
      case 'e': /* save as eps */
        if ( freq == NULL ) /* there was never an acquisition */
	  break;
	strcpy( message, READY );
	int_ask_str( "Basename for file", base_name );
	if ( ! base_name[0] )
	{
	  strcpy( message, "Basename empty." );
	  break;
	}
	fname = make_filename( arg, base_name, (c=='p')?"ps":"eps" ); 
	write_to_gnuplot( gp, arg, freq, abslt, phase, fname, 
	                  (c=='p')?OUT_PS:OUT_EPS );
	fflush( gp );
	free( fname );
        break;
      case 'g': /* save Gnuplot command file */
        if ( freq == NULL ) /* there was never an acquisition */
	  break;
	strcpy( message, READY );
	int_ask_str( "Basename for file", base_name );
	if ( ! base_name[0] )
	{
	  strcpy( message, "Basename empty." );
	  break;
	}
	fname = make_filename( arg, base_name, "eps" ); 
	fname2 = make_filename( arg, base_name, "plt" ); 
	o = fopen( fname2, "w" );
	write_to_gnuplot( o, arg, freq, abslt, phase, fname, OUT_EPS );
	fclose( o );
	free( fname );
	free( fname2 );
        break;
      case 'd': /* save Data file */
        if ( freq == NULL ) /* there was never an acquisition */
	  break;
	strcpy( message, READY );
	int_ask_str( "Basename for file", base_name );
	if ( ! base_name[0] )
	{
	  strcpy( message, "Basename empty." );
	  break;
	}
	write_dat( arg, freq, abslt, phase, base_name );
	break;
      case 't': /* prinT */
        if ( freq == NULL ) /* there was never an acquisition */
	  break;
	strcpy( message, READY );
	int_ask_str( "Printer command", &printer[1] );
	if ( ! printer[1] )
	{
	  strcpy( message, "Printer command empty." );
	  break;
	}
	write_to_gnuplot( gp, arg, freq, abslt, phase, printer, OUT_PS );
	fflush( gp );
        break;
      case 'q': /* Quit. */
        break;
      default: /* unknown letter */
        sprintf( message, "Unknown command '%c'.", c );
    }
  }
  while ( c != 'q' );
  close_gnuplot( gp );
  free( freq );
  free( abslt );
  free( phase );
}


/* ---------------------------------------------------------------------- 
 * main program and helpers
 * ---------------------------------------------------------------------- */


/** Print help to stdout and exit with 0 */
void help()
{
  puts( "HMsweep" );
  puts( "(c) Lars Stollenwerk, 2007 - 2016" );
  puts( "" );
  puts( "hmsweep -h            print this help" );
  puts( "hmsweep -v            print version and copyright" );
  puts( "hmsweep -v <serial>   print firmware version for scope\n" );
  puts( "hmsweep -i [options]  enter interactive mode" );
  puts( "hmsweep [options] name mode fmin fmax [fsteps] " );
  puts( "" );
  puts( "name                  basename for outputfiles." );
  puts( "                      Files are named 'name_mode_fmin_fmax.ending'." );
  puts( "mode                  'abs' - only absolute of CH 1 is captured" );
  puts( "                      'bode' - amplitude and phase relative to CH 2" );
  puts( "                      'imp' - impedance measurement; " );
  puts( "                              U = CH 1, I = -CH 2 / R" );
  puts( "fmin                  minimal frequency" );
  puts( "fmax                  maximal frequency" );
  puts( "fsteps                number of steps, default are halvetones" );
  puts( "" );
  puts( "If fmax < fmin, the sweep goes downwards." );
  puts( "" );
  puts( "options:" );
  printf( "  -s serial           serial port for scope, default '%s'\n", 
          SERIAL_PORT_OSC );
  printf( "  -a alsa             use sound card for signal generation using \n"
          "                      alsa device, default '%s'\n"
	  "                      Frequency range: 100 Hz to 22050 Hz\n", 
	  ALSA_DEVICE );
  puts( "  -S serial           use function generator at serial device for\n"
        "                      signal generation\n"
	"                      Frequency range: 100 mHz to 15 MHz" );
  puts( "  -r R                resistor for current measurement in impedance "
	"mode" );
  puts( "" );
  puts( "At least one of the output modes must be given:" );
  puts( "  -p                  PS file" );
  puts( "  -e                  EPS file" );
  puts( "  -x                  X11 output" );
  puts( "  -g                  gnuplot file .plt" );
  puts( "  -d                  data file" );
  puts( "" );
  puts( "fmin, fmax and R may be given with SI-prefix (e.g.: 1k = 1000)" );
  exit( 0 );
}


/** Print program version and copyright and exit with 0 */
void copyleft()
{
  puts( "HM sweep - measuring transfer functions and impedances in" );
  puts( "           dependence of the signal frequency" );
  puts( "           using the soundcard and an HAMEG oscilloscope" );
  puts( "" );
  printf( "Version %s\n", VERSION );
  puts( "Copyright (C) 2007 - 2008 Lars Stollenwerk" );
  puts( "" );
  puts( "This program is free software; you can redistribute it and/or" );
  puts( "modify it under the terms of the GNU General Public License as" );
  puts( "published by the Free Software Foundation; either version 2 of" );
  puts( "the License, or (at your option) any later version." );
  puts( "" );
  puts( "This program is distributed in the hope that it will be" );
  puts( "useful, but WITHOUT ANY WARRANTY; without even the implied" );
  puts( "warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR" );
  puts( "PURPOSE. See the GNU General Public License for more" );
  puts( "details." );
  puts( "" );
  puts( "You should have received a copy of the GNU General" );
  puts( "Public License along with this program; if not, write" );
  puts( "to the Free Software Foundation, Inc., 51 Franklin St," );
  puts( "Fifth Floor, Boston, MA 02110, USA" );
  puts( "" );
  puts( "Contact the author via" );
  puts( "LSto@opppf.de" );
  puts( "http://opppf.de/HMsweep/" );
  exit( 0 );
}


/** Print scope version to stderr */
void version( const char * port )
{
  OSC * osc;

  osc = osc_open( port );
  if ( osc->error[0] ) 
  {
    fprintf( stderr, "Error: %s\n", osc->error );
    exit( 1 );
  }
  else
    fprintf( stderr, "Oscilloscope Version: %s\n", osc->version );
  osc_close( osc );
  exit( 0 );
}


/** read commandline args */
void read_args( int argc, char ** argv, arg_t * arg, arg_sweep_t * arg_sweep )
{
  int i;
  char msg[256];

  /* set defaults */
  arg->help = 0;
  arg->version = 0;
  arg->interactive = 0;
  arg->output_mode = 0;
  arg_sweep->serial = SERIAL_PORT_OSC;
  arg_sweep->alsa = ALSA_DEVICE;
  arg_sweep->func = NULL;
  arg_sweep->R = R_DEFAULT;
  arg_sweep->genFmin = 10;
  arg_sweep->genFmax = 22050;
  /* options, beginning with - */
  i = 0;
  while ( ++i < argc && argv[i][0] == '-' )
  {
    switch ( argv[i][1] )
    {
      case 'h': arg->help = 1; break;
      case 'v':
	arg->version = 1;
	if ( ++i < argc )
	{
	    arg->version = 2;
	    arg_sweep->serial = argv[i]; 
	}
	break;
      case 'i': arg->interactive = 1; break;
      case 's': 
        if ( ++i == argc ) error( "Name a serial port." );
        arg_sweep->serial = argv[i]; 
	break;
      case 'a': 
        if ( ++i == argc ) error( "Name an alsa device." );
        arg_sweep->alsa = argv[i]; 
        arg_sweep->func = NULL;
	arg_sweep->genFmin = 10;
	arg_sweep->genFmax = 22050;
	break;
      case 'S': 
        if ( ++i == argc ) error( "Name a serial port." );
        arg_sweep->alsa = NULL; 
        arg_sweep->func = argv[i];
	arg_sweep->genFmin = 100e-6;
	arg_sweep->genFmax = 15e6;
	break;
      case 'r': 
        if ( ++i == argc ) error( "Provide a value for R." );
        arg_sweep->R = scan_si( argv[i] ); 
	break;
      case 'p': arg->output_mode += OUT_PS; break;
      case 'e': arg->output_mode += OUT_EPS; break;
      case 'x': arg->output_mode += OUT_X11; break;
      case 'g': arg->output_mode += OUT_PLT; break;
      case 'd': arg->output_mode += OUT_DAT; break;
      default: 
        error( "Unknown option." );
    }
  }
  i--;
  /* no further operation on pure informative runs */
  if ( arg->version ) 
    return;
  if ( arg->help ) 
    return;
  /* other arguments */
  if ( arg->interactive )
  {
    arg->name = NULL;
    arg_sweep->mode = 'b';
    arg_sweep->fmin = 100;
    arg_sweep->fmax = 10000;
    arg_sweep->fsteps = halftones( arg_sweep->fmax/arg_sweep->fmin );
    arg_sweep->fstep = HALFTONE;
  }
  else
  {
    if ( ++i == argc ) error( "Provide a filename." );
    arg->name = argv[i];
    if ( ++i == argc ) error( "Provide an operation mode." );
    arg_sweep->mode = argv[i][0];
    if ( ++i == argc ) error( "Provide a start frequency (fmin)." );
    arg_sweep->fmin = scan_si( argv[i] );
    if ( ++i == argc ) error( "Provide a stop frequency (fmax)." );
    arg_sweep->fmax = scan_si( argv[i] );
    if ( ++i == argc )
    {
      arg_sweep->fsteps = halftones( arg_sweep->fmax/arg_sweep->fmin );
      arg_sweep->fstep = calc_fstep( arg_sweep->fmin, arg_sweep->fmax, 
                                     arg_sweep->fsteps ); 
    }
    else
    {
      arg_sweep->fsteps = abs( atoi( argv[i] ) );
      arg_sweep->fstep = calc_fstep( arg_sweep->fmin, arg_sweep->fmax, 
                                     arg_sweep->fsteps ); 
    }
  }
  /* check consistency */
  if ( arg_sweep->fmin < arg_sweep->genFmin ) 
  {
    snprintf( msg, 256, "Choose fmin >= %g Hz.", arg_sweep->genFmin );
    error( msg );
  }
  if ( arg_sweep->fmax > arg_sweep->genFmax ) 
  {
    snprintf( msg, 256, "Choose fmax <= %g Hz.", arg_sweep->genFmax );
    error( msg );
  }
  if ( arg_sweep->fmin == arg_sweep->fmax )
    error( "f_min and f_max must differ." );
  if ( fabs( arg_sweep->fstep - 1 ) < 0.002 )
    fprintf( stderr, 
      "%i steps from %.1f Hz to %.1f Hz are pretty much. Think about it!\n", 
      arg_sweep->fsteps, arg_sweep->fmin, arg_sweep->fmax );
  if ( strchr( "abi", arg_sweep->mode ) == NULL )
    error( "Unknown sweep mode" );
  if ( ! arg->interactive && ! arg->output_mode )
    error( "What kind of output do you like?" );
}


/** Print summary of arguments */
void print_args( const arg_t * arg, const arg_sweep_t * arg_sweep ) 
{
  fputs( "--------------------------\n", stderr );
  fprintf( stderr, "Scope at:     %12s\n", arg_sweep->serial );
  if ( arg_sweep->alsa != NULL )
    fprintf( stderr, "Alsa device:  %12s\n", arg_sweep->alsa );
  else
    fprintf( stderr, "Func gen at:  %12s\n", arg_sweep->func );
  fprintf( stderr, "fmin [Hz]:    %12.1f\n", arg_sweep->fmin );
  fprintf( stderr, "fmax [Hz]:    %12.1f\n", arg_sweep->fmax );
  fprintf( stderr, "steps:        %12i\n", arg_sweep->fsteps  );
  fprintf( stderr, "factor:       %12.5f\n", arg_sweep->fstep  );
  fprintf( stderr, "mode:         %12c\n", arg_sweep->mode );
  if ( arg_sweep->mode == 'i' )
    fprintf( stderr, "R [Ohm]:      %12g\n", arg_sweep->R );
  if ( arg->output_mode & ( OUT_PS + OUT_EPS + OUT_PLT + OUT_DAT ) )
    fprintf( stderr, "Basename:     %12s\n", arg->name );
  if ( arg->output_mode & OUT_PS )
    fputs( "Writing                 PS\n", stderr );
  if ( arg->output_mode & OUT_EPS )
    fputs( "Writing                EPS\n", stderr );
  if ( arg->output_mode & OUT_PLT )
    fputs( "Writing                PLT\n", stderr );
  if ( arg->output_mode & OUT_DAT )
    fputs( "Writing                DAT\n", stderr );
  if ( arg->output_mode & OUT_X11 )
    fputs( "Showing                X11\n", stderr );
  fputs( "--------------------------\n", stderr );
}


/**
 * Hauptprogramm
 */
int main( int argc, char ** argv ) 
{
  arg_t arg;
  arg_sweep_t arg_sweep;
  double * freq;
  double * abslt;
  double * phase;
  char * fname;
  char * fname2;
  FILE * o;

  read_args( argc, argv, &arg, &arg_sweep );
  if ( arg.help ) help();
  if ( arg.version == 1 ) copyleft();
  if ( arg.version == 2 ) version( arg_sweep.serial );
  if ( arg.interactive )
    interactive( &arg_sweep );
  else
  {
    print_args( &arg, &arg_sweep );
    block_sig_int();
    puts( "Sweeping. Abort with ^C." );
    if ( sweep( &arg_sweep, &freq, &abslt, &phase ) != 0 )
      exit( 0 );
    /* output using gnuplot */
    if ( arg.output_mode & ( OUT_PS | OUT_EPS | OUT_X11 ) )
      o = open_gnuplot( 1 );
    if ( arg.output_mode & OUT_PS )
    {
      fname = make_filename( &arg_sweep, arg.name, "ps" ); 
      write_to_gnuplot( o, &arg_sweep, freq, abslt, phase, fname, OUT_PS );
      free( fname );
    }
    if ( arg.output_mode & OUT_EPS )
    {
      fname = make_filename( &arg_sweep, arg.name, "eps" ); 
      write_to_gnuplot( o, &arg_sweep, freq, abslt, phase, fname, OUT_EPS );
      free( fname );
    }
    if ( arg.output_mode & OUT_X11 )
    {
      write_to_gnuplot( o, &arg_sweep, freq, abslt, phase, NULL, OUT_X11 );
    }
    if ( arg.output_mode & ( OUT_PS | OUT_EPS | OUT_X11 ) )
      close_gnuplot( o );
    /* output not using gnuplot */
    if ( arg.output_mode & OUT_PLT )
    {
      fname = make_filename( &arg_sweep, arg.name, "eps" ); 
      fname2 = make_filename( &arg_sweep, arg.name, "plt" ); 
      o = fopen( fname2, "w" );
      write_to_gnuplot( o, &arg_sweep, freq, abslt, phase, fname, OUT_EPS );
      fclose( o );
      free( fname );
      free( fname2 );
    }
    if ( arg.output_mode & OUT_DAT )
      write_dat( &arg_sweep, freq, abslt, phase, arg.name );
    /* clean up */
    free( freq );
    free( abslt );
    free( phase );
  }
  return 0;
}


/* vim: set fenc=latin1 fdm=expr foldexpr=getline(v\:lnum)=~'^{'?'>1'\:getline(v\:lnum)=~'^}'?'<1'\:'-1' : */
