#include "lib232.h"
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>

/* Set bit y in variable x.  */
#define SetBit(x,y) ((x) |= (y))

/* Clear bit y in variable x. */
#define ClrBit(x,y) ((x) &= ~(y))

int open232( const char * tty, int baud, const char * frame, int hs, int mode )
{
    int fd;
    struct termios tio;
    speed_t speed;
    int save_errno;

    fd = open( tty, O_RDWR | O_NOCTTY );
    if ( fd == -1 )
	goto IOerror;
    if ( tcgetattr( fd, &tio ) == -1 )
	goto IOerror;
    /* tio.c_iflag */
    SetBit( tio.c_iflag, IGNBRK ); /* ign. break; BRKINT irrelevant */
    SetBit( tio.c_iflag, IGNPAR ); /* ign. framing errors; PARMRK irrelevant */
    ClrBit( tio.c_iflag, INPCK );  /* input parity check */
    ClrBit( tio.c_iflag, ISTRIP ); /* strip 8th bit */
    ClrBit( tio.c_iflag, INLCR );  /* \n -> \r */
    ClrBit( tio.c_iflag, IGNCR );  /* ign. \r */
    ClrBit( tio.c_iflag, IXANY );  /* any char restarts output */
    /* non POSIX */
    #ifdef IUCLC
    ClrBit( tio.c_iflag, IUCLC );  /* to lowercase */
    #endif
    #ifdef IMAXBEL
    ClrBit( tio.c_iflag, IMAXBEL );/* ring bell on full input queue */
    #endif
    #ifdef IUTF8
    ClrBit( tio.c_iflag, IUTF8 );  /* input is UTF-8 */
    #endif
    /* tio.c_oflag */
    ClrBit( tio.c_oflag, ISTRIP ); /* enable output processing (clr=raw out) */
    /* tio.c_cflag */
    ClrBit( tio.c_cflag, CSIZE );  /* clear character size */
    SetBit( tio.c_cflag, CREAD );  /* enable receiver */
    SetBit( tio.c_cflag, HUPCL );  /* drop DTR on close */
    SetBit( tio.c_cflag, CLOCAL ); /* ign. modem control lines */
    /* tio.c_lflag */
    ClrBit( tio.c_lflag, ISIG );   /* enable signals INTR QUIT SUSP DSUSP */
    ClrBit( tio.c_lflag, ECHO );   /* echo input chars */
    ClrBit( tio.c_lflag, ECHOE );  /* on ICANON, enable ERASE */
    ClrBit( tio.c_lflag, ECHOK );  /* on ICANON, enable KILL */
    ClrBit( tio.c_lflag, ECHONL ); /* on ICANON, echo NL  */
    ClrBit( tio.c_lflag, NOFLSH ); /* don't flush on SIGINT SIGQUIT SIGSUSP */
    ClrBit( tio.c_lflag, TOSTOP ); /* SIGTTOU on write to controlling term */
    /* non POSIX */
    #ifdef XCASE
    ClrBit( tio.c_lflag, XCASE );  /* uppercase terminal */
    #endif
    #ifdef ECHOCTL
    ClrBit( tio.c_lflag, ECHOCTL );/* echo control chars as ^X */
    #endif
    #ifdef ECHOPRT
    ClrBit( tio.c_lflag, ECHOPRT );/* print erased chars */
    #endif
    #ifdef ECHOKE
    ClrBit( tio.c_lflag, ECHOKE ); /* echo KILL */
    #endif
    #ifdef DEFECHO
    ClrBit( tio.c_lflag, DEFECHO );/* echo only while reading */
    #endif
    #ifdef FLUSHO
    ClrBit( tio.c_lflag, FLUSHO ); /* output ist flusched */
    #endif
    #ifdef PENDIN
    ClrBit( tio.c_lflag, PENDIN ); /* reprint chars in queue on next read */
    #endif
    #ifdef IEXTEN
    ClrBit( tio.c_lflag, IEXTEN ); /* enable extende processing */
    #endif

    /* set frame */
    if ( frame == NULL || strlen(frame) != 3 )
    {
	errno = EINVAL;
	goto IOerror;
    }
    switch ( frame[0] )
    {
        case '5': SetBit( tio.c_cflag, CS5 ); break;
        case '6': SetBit( tio.c_cflag, CS6 ); break;
        case '7': SetBit( tio.c_cflag, CS7 ); break;
        case '8': SetBit( tio.c_cflag, CS8 ); break;
        default: 
	    errno = EINVAL;
	    goto IOerror;
    }
    switch ( frame[1] )
    {
        case 'n': case 'N':
	    ClrBit( tio.c_cflag, PARENB );
	    break;
        case 'o': case 'O':
	    SetBit( tio.c_cflag, PARENB );
	    SetBit( tio.c_cflag, PARODD );
	    break;
        case 'e': case 'E':
	    SetBit( tio.c_cflag, PARENB );
	    ClrBit( tio.c_cflag, PARODD );
	    break;
        default: 
	    errno = EINVAL;
	    goto IOerror;
    }
    switch ( frame[2] )
    {
        case '1' : 
	    ClrBit( tio.c_cflag, CSTOPB ); /* one stop bit */
	    break;
        case '2' : 
	    SetBit( tio.c_cflag, CSTOPB ); /* two stop bits */
	    break;
        default: 
	    errno = EINVAL;
	    goto IOerror;
    }
    /* set handshake */
    switch ( hs )
    {
        case NONE: 
	    ClrBit( tio.c_iflag, IXON );   /* XON/XOFF flow contorl, output */
	    ClrBit( tio.c_iflag, IXOFF );  /* XON/XOFF flow contorl, input */
	    #ifdef CRTSCTS
	    ClrBit( tio.c_cflag, CRTSCTS );/* RTS/CTS handshake, non POSIX */
	    #endif
	    break;
        case RTSCTS:
	    ClrBit( tio.c_iflag, IXON );   /* XON/XOFF flow contorl, output */
	    ClrBit( tio.c_iflag, IXOFF );  /* XON/XOFF flow contorl, input */
	    #ifdef CRTSCTS
	    SetBit( tio.c_cflag, CRTSCTS );/* RTS/CTS handshake, non POSIX */
	    #else
	    fputs( "open232: Caution: non POSIX RTS/CTS hanshake requested "
		   "but not available.\n", stderr );
	    #endif
	    break;
        case XONXOFF:
	    SetBit( tio.c_iflag, IXON );   /* XON/XOFF flow contorl, output */
	    SetBit( tio.c_iflag, IXOFF );  /* XON/XOFF flow contorl, input */
	    #ifdef CRTSCTS
	    ClrBit( tio.c_cflag, CRTSCTS );/* RTS/CTS handshake, non POSIX */
	    #endif
	    break;
        default: 
	    errno = EINVAL;
	    goto IOerror;
    }
    /* set mode */
    if ( mode )
    {
        SetBit( tio.c_iflag, ICRNL );  /* \r -> \n */
        SetBit( tio.c_lflag, ICANON ); /* enable canonical mode */
    }
    else
    {
        ClrBit( tio.c_iflag, ICRNL );
        ClrBit( tio.c_lflag, ICANON );
        tio.c_cc[VMIN] = 1;
        tio.c_cc[VTIME] = 0;
    }
    /* set baud rate */
    switch ( baud  )
    {
        case 50: speed = B50; break;
        case 75: speed = B75; break;
        case 110: speed = B110; break;
        case 134: speed = B134; break;
        case 150: speed = B150; break;
        case 200: speed = B200; break;
        case 300: speed = B300; break;
        case 600: speed = B600; break;
        case 1200: speed = B1200; break;
        case 1800: speed = B1800; break;
        case 2400: speed = B2400; break;
        case 4800: speed = B4800; break;
        case 9600: speed = B9600; break;
        case 19200: speed = B19200; break;
        case 38400: speed = B38400; break;
        case 57600: speed = B57600; break;
        case 115200: speed = B115200; break;
        case 230400: speed = B230400; break;
        default: 
	    errno = EINVAL;
	    goto IOerror;
    }
    if ( cfsetispeed( &tio, speed ) == -1 )
	goto IOerror;
    if ( cfsetospeed( &tio, speed ) == -1 )
	goto IOerror;
    /* set termios */
    if ( tcflush( fd, TCIOFLUSH ) == -1 )
	goto IOerror;
    if ( tcsetattr( fd, TCSANOW, &tio ) == -1 )
	goto IOerror;
    /* return sucessfully */
    return fd;
IOerror:
    save_errno = errno;
    if ( fd != -1 ) close( fd );
    errno = save_errno;
    return -1;
}


int print_rs232_state( int fd )
{
    int status;

    if ( ioctl( fd, TIOCMGET, &status ) == -1 )
	return -1;
    fprintf( stderr, "| >DTR <DSR | >RTS <CTS | >CD  <RI  |\n" );
    fprintf( stderr, "|  %c    %c   |  %c    %c   |  %c    %c   |\n",
  	     (status & TIOCM_DTR )?'1':'.',
  	     (status & TIOCM_DSR )?'1':'.',
  	     (status & TIOCM_RTS )?'1':'.',
  	     (status & TIOCM_CTS )?'1':'.',
  	     (status & TIOCM_CD  )?'1':'.',
  	     (status & TIOCM_RI  )?'1':'.' );
    return 0;
}

